/*
 * Block descriptor functions
 *
 * Copyright (C) 2011-2012, Joachim Metz <jbmetz@users.sourceforge.net>
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <memory.h>
#include <types.h>

#include "libvshadow_array_type.h"
#include "libvshadow_block_descriptor.h"
#include "libvshadow_btree_type.h"
#include "libvshadow_debug.h"
#include "libvshadow_definitions.h"
#include "libvshadow_libcerror.h"
#include "libvshadow_libcnotify.h"

#include "vshadow_store.h"

const uint8_t vshadow_empty_store_block_list_entry[ 32 ] = {
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

/* Initialize a block descriptor
 * Make sure the value block_descriptor is pointing to is set to NULL
 * Returns 1 if successful or -1 on error
 */
int libvshadow_block_descriptor_initialize(
     libvshadow_block_descriptor_t **block_descriptor,
     libcerror_error_t **error )
{
	static char *function = "libvshadow_block_descriptor_initialize";

	if( block_descriptor == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid block descriptor.",
		 function );

		return( -1 );
	}
	if( *block_descriptor != NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid block descriptor value already set.",
		 function );

		return( -1 );
	}
	*block_descriptor = memory_allocate_structure(
	                     libvshadow_block_descriptor_t );

	if( block_descriptor == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_MEMORY,
		 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
		 "%s: unable to create block descriptor.",
		 function );

		goto on_error;
	}
	if( memory_set(
	     *block_descriptor,
	     0,
	     sizeof( libvshadow_block_descriptor_t ) ) == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_MEMORY,
		 LIBCERROR_MEMORY_ERROR_SET_FAILED,
		 "%s: unable to clear block descriptor.",
		 function );

		goto on_error;
	}
	if( libvshadow_array_initialize(
	     &( ( *block_descriptor )->overlays_array ),
	     0,
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create overlays array.",
		 function );

		goto on_error;
	}
	( *block_descriptor )->index = -1;

	return( 1 );

on_error:
	if( *block_descriptor != NULL )
	{
		memory_free(
		 *block_descriptor );

		*block_descriptor = NULL;
	}
	return( -1 );
}

/* Frees a block descriptor
 * Returns 1 if successful or -1 on error
 */
int libvshadow_block_descriptor_free(
     libvshadow_block_descriptor_t **block_descriptor,
     libcerror_error_t **error )
{
	static char *function = "libvshadow_block_descriptor_free";
	int result            = 1;

	if( block_descriptor == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid block descriptor.",
		 function );

		return( -1 );
	}
	if( *block_descriptor != NULL )
	{
/* TODO
		if( libvshadow_array_free(
		     &( ( *block_descriptor )->overlays_array ),
		     (int (*)(intptr_t **, libcerror_error_t **)) &libvshadow_block_descriptor_free,
		     error ) != 1 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_FINALIZE_FAILED,
			 "%s: unable to free overlays array.",
			 function );

			result = -1;
		}
*/
		memory_free(
		 *block_descriptor );

		*block_descriptor = NULL;
	}
	return( result );
}

/* Compares 2 block descriptors by the original offset
 * Returns LIBVSHADOW_ARRAY_COMPARE_LESS, LIBVSHADOW_ARRAY_COMPARE_EQUAL, LIBVSHADOW_ARRAY_COMPARE_GREATER if successful or -1 on error
 */
int libvshadow_block_descriptor_compare_by_original_offset(
     libvshadow_block_descriptor_t *first_block_descriptor,
     libvshadow_block_descriptor_t *second_block_descriptor,
     libcerror_error_t **error )
{
	static char *function = "libvshadow_block_descriptor_compare_by_original_offset";

	if( first_block_descriptor == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid first block descriptor.",
		 function );

		return( -1 );
	}
	if( second_block_descriptor == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid second block descriptor.",
		 function );

		return( -1 );
	}
	if( first_block_descriptor->original_offset < second_block_descriptor->original_offset )
	{
		return( LIBVSHADOW_BTREE_NODE_COMPARE_LESS );
	}
	else if( first_block_descriptor->original_offset > second_block_descriptor->original_offset )
	{
		return( LIBVSHADOW_BTREE_NODE_COMPARE_GREATER );
	}
	return( LIBVSHADOW_BTREE_NODE_COMPARE_EQUAL );
}

/* Compares 2 block descriptors by the relative offset
 * Returns LIBVSHADOW_ARRAY_COMPARE_LESS, LIBVSHADOW_ARRAY_COMPARE_EQUAL, LIBVSHADOW_ARRAY_COMPARE_GREATER if successful or -1 on error
 */
int libvshadow_block_descriptor_compare_by_relative_offset(
     libvshadow_block_descriptor_t *first_block_descriptor,
     libvshadow_block_descriptor_t *second_block_descriptor,
     libcerror_error_t **error )
{
	static char *function = "libvshadow_block_descriptor_compare_by_relative_offset";

	if( first_block_descriptor == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid first block descriptor.",
		 function );

		return( -1 );
	}
	if( second_block_descriptor == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid second block descriptor.",
		 function );

		return( -1 );
	}
	if( first_block_descriptor->relative_offset < second_block_descriptor->relative_offset )
	{
		return( LIBVSHADOW_BTREE_NODE_COMPARE_LESS );
	}
	else if( first_block_descriptor->relative_offset > second_block_descriptor->relative_offset )
	{
		return( LIBVSHADOW_BTREE_NODE_COMPARE_GREATER );
	}
	return( LIBVSHADOW_BTREE_NODE_COMPARE_EQUAL );
}

/* Reads the block descriptor
 * Returns 1 if successful, 0 if block list entry is empty or -1 on error
 */
int libvshadow_block_descriptor_read(
     libvshadow_block_descriptor_t *block_descriptor,
     const uint8_t *block_list_entry_data,
     size_t block_list_entry_size,
     libcerror_error_t **error )
{
	static char *function      = "libvshadow_block_descriptor_read";
	uint32_t allocation_bitmap = 0;

	if( block_descriptor == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid block descriptor.",
		 function );

		return( -1 );
	}
	if( block_list_entry_data == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid block list entry data.",
		 function );

		return( -1 );
	}
	if( block_list_entry_size > (size_t) SSIZE_MAX )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: invalid block list entry size value exceeds maximum.",
		 function );

		return( -1 );
	}
	if( memory_compare(
	     block_list_entry_data,
             vshadow_empty_store_block_list_entry,
             sizeof( vshadow_store_block_list_entry_t ) ) == 0 )
	{
		return( 0 );
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libcnotify_verbose != 0 )
	{
		libcnotify_printf(
		 "%s: block list entry data:\n",
		 function );
		libcnotify_print_data(
		 block_list_entry_data,
		 sizeof( vshadow_store_block_list_entry_t ),
		 0 );
	}
#endif
	byte_stream_copy_to_uint64_little_endian(
	 ( (vshadow_store_block_list_entry_t *) block_list_entry_data )->original_offset,
	 block_descriptor->original_offset );

	byte_stream_copy_to_uint64_little_endian(
	 ( (vshadow_store_block_list_entry_t *) block_list_entry_data )->relative_offset,
	 block_descriptor->relative_offset );

	byte_stream_copy_to_uint64_little_endian(
	 ( (vshadow_store_block_list_entry_t *) block_list_entry_data )->offset,
	 block_descriptor->offset );

	byte_stream_copy_to_uint32_little_endian(
	 ( (vshadow_store_block_list_entry_t *) block_list_entry_data )->flags,
	 block_descriptor->flags );

	byte_stream_copy_to_uint32_little_endian(
	 ( (vshadow_store_block_list_entry_t *) block_list_entry_data )->allocation_bitmap,
	 allocation_bitmap );

#if defined( HAVE_DEBUG_OUTPUT )
	if( libcnotify_verbose != 0 )
	{
		libcnotify_printf(
		 "%s: original offset\t\t\t: 0x%08" PRIx64 "\n",
		 function,
		 block_descriptor->original_offset );

		libcnotify_printf(
		 "%s: relative offset\t\t\t: 0x%08" PRIx64 "\n",
		 function,
		 block_descriptor->relative_offset );

		libcnotify_printf(
		 "%s: offset\t\t\t\t: 0x%08" PRIx64 "\n",
		 function,
		 block_descriptor->offset );

		libcnotify_printf(
		 "%s: flags\t\t\t\t\t: 0x%08" PRIx32 "\n",
		 function,
		 block_descriptor->flags );

		libcnotify_printf(
		 "%s: allocation bitmap\t\t\t: 0x%08" PRIx32 "\n",
		 function,
		 allocation_bitmap );

		libcnotify_printf(
		 "\n" );
	}
#endif
	if( ( block_descriptor->flags & LIBVSHADOW_BLOCK_DESCRIPTOR_FLAGS_IS_FORWARDER ) != 0 )
	{
		if( block_descriptor->offset != 0 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
			 "%s: unsupported offset - value not zero.",
			 function );

			return( -1 );
		}
	}
	if( ( block_descriptor->flags & LIBVSHADOW_BLOCK_DESCRIPTOR_FLAGS_IS_OVERLAY ) != 0 )
	{
		while( ( allocation_bitmap & 0x01 ) == 0 )
		{
			block_descriptor->block_offset += 512;

			allocation_bitmap >>= 1;
		}
		block_descriptor->block_size = 0;

		while( ( allocation_bitmap & 0x01 ) != 0 )
		{
			block_descriptor->block_size += 512;

			allocation_bitmap >>= 1;
		}
		if( allocation_bitmap != 0 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_UNSUPPORTED_VALUE,
			 "%s: unsupported allocation bitmap - not a continuous range.",
			 function );

			return( -1 );
		}
	}
	else
	{
		block_descriptor->block_size = 0x4000;
	}
	return( 1 );
}

