/*
 * Block tree functions
 *
 * Copyright (c) 2011-2012, Joachim Metz <jbmetz@users.sourceforge.net>
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <types.h>

#include "libvshadow_block_descriptor.h"
#include "libvshadow_btree_type.h"
#include "libvshadow_definitions.h"
#include "libvshadow_libcerror.h"
#include "libvshadow_libcnotify.h"
#include "libvshadow_list_type.h"
#include "libvshadow_tree_type.h"

/* Retrieves the block descriptor for the specific orignal offset
 * Returns 1 if successful, 0 if no such value or -1 on error
 */
int libvshadow_block_tree_node_get_descriptor_by_original_offset(
     libvshadow_tree_node_t *node,
     off64_t original_offset,
     libvshadow_block_descriptor_t **block_descriptor,
     libcerror_error_t **error )
{
	libvshadow_list_element_t *values_list_element = NULL;
	libvshadow_tree_node_t *sub_node               = NULL;
	static char *function                          = "libvshadow_block_tree_node_get_descriptor_by_original_offset";
	int number_of_values_list_elements             = 0;
	int result                                     = 0;
	int sub_node_index                             = 0;
	int values_list_element_index                  = 0;

	if( node == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid node.",
		 function );

		return( -1 );
	}
	if( block_descriptor == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid block descriptor.",
		 function );

		return( -1 );
	}
	if( node->value != NULL )
	{
		if( libvshadow_list_get_number_of_elements(
		     (libvshadow_list_t *) node->value,
		     &number_of_values_list_elements,
		     error ) != 1 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve number of values list elements.",
			 function );

			return( -1 );
		}
	}
	if( number_of_values_list_elements == 0 )
	{
		return( 0 );
	}
	if( libvshadow_list_get_element_by_index(
	     (libvshadow_list_t *) node->value,
	     0,
	     &values_list_element,
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve values list element: 0.",
		 function );

		return( -1 );
	}
	if( node->number_of_sub_nodes != 0 )
	{
		if( ( number_of_values_list_elements + 1 ) != node->number_of_sub_nodes )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBCERROR_ARGUMENT_ERROR_VALUE_OUT_OF_BOUNDS,
			 "%s: invalid number of values list elements value out of bounds.",
			 function );

			return( -1 );
		}
		sub_node = node->first_sub_node;
	}
	for( values_list_element_index = 0;
	     values_list_element_index < number_of_values_list_elements;
	     values_list_element_index++ )
	{
		if( values_list_element == NULL )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: missing values list element: %d.",
			 function,
			 values_list_element_index );

			return( -1 );
		}
		if( values_list_element->value == NULL )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid values list element: %d - missing value.",
			 function,
			 values_list_element_index );

			return( -1 );
		}
		*block_descriptor = (libvshadow_block_descriptor_t *) values_list_element->value;

		if( ( original_offset >= ( *block_descriptor )->original_offset )
		 && ( original_offset < ( ( *block_descriptor )->original_offset + 0x4000 ) ) )
		{
			return( 1 );
		}
		else if( original_offset < ( *block_descriptor )->original_offset )
		{
			break;
		}
		values_list_element = values_list_element->next_element;

		if( node->number_of_sub_nodes != 0 )
		{
			if( sub_node == NULL )
			{
				libcerror_error_set(
				 error,
				 LIBCERROR_ERROR_DOMAIN_RUNTIME,
				 LIBCERROR_RUNTIME_ERROR_VALUE_MISSING,
				 "%s: missing sub node: %d.",
				 function,
				 sub_node_index );

				return( -1 );
			}
			sub_node = sub_node->next_node;

			sub_node_index++;
		}
	}
	*block_descriptor = NULL;

	if( node->number_of_sub_nodes == 0 )
	{
		return( 0 );
	}
	result = libvshadow_block_tree_node_get_descriptor_by_original_offset(
		  sub_node,
		  original_offset,
		  block_descriptor,
		  error );

	if( result == -1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve block descriptor in sub node: %d.",
		 function,
		 sub_node_index );

		return( -1 );
	}
	return( result );
}

/* Retrieves the block descriptor for the specific relative offset
 * Returns 1 if successful, 0 if no such value or -1 on error
 */
int libvshadow_block_tree_node_get_descriptor_by_relative_offset(
     libvshadow_tree_node_t *node,
     off64_t relative_offset,
     libvshadow_block_descriptor_t **block_descriptor,
     libcerror_error_t **error )
{
	libvshadow_list_element_t *values_list_element = NULL;
	libvshadow_tree_node_t *sub_node               = NULL;
	static char *function                          = "libvshadow_block_tree_node_get_descriptor_by_relative_offset";
	int number_of_values_list_elements             = 0;
	int result                                     = 0;
	int sub_node_index                             = 0;
	int values_list_element_index                  = 0;

	if( node == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid node.",
		 function );

		return( -1 );
	}
	if( block_descriptor == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid block descriptor.",
		 function );

		return( -1 );
	}
	if( node->value != NULL )
	{
		if( libvshadow_list_get_number_of_elements(
		     (libvshadow_list_t *) node->value,
		     &number_of_values_list_elements,
		     error ) != 1 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve number of values list elements.",
			 function );

			return( -1 );
		}
	}
	if( number_of_values_list_elements == 0 )
	{
		return( 0 );
	}
	if( libvshadow_list_get_element_by_index(
	     (libvshadow_list_t *) node->value,
	     0,
	     &values_list_element,
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve values list element: 0.",
		 function );

		return( -1 );
	}
	if( node->number_of_sub_nodes != 0 )
	{
		if( ( number_of_values_list_elements + 1 ) != node->number_of_sub_nodes )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBCERROR_ARGUMENT_ERROR_VALUE_OUT_OF_BOUNDS,
			 "%s: invalid number of values list elements value out of bounds.",
			 function );

			return( -1 );
		}
		sub_node = node->first_sub_node;
	}
	for( values_list_element_index = 0;
	     values_list_element_index < number_of_values_list_elements;
	     values_list_element_index++ )
	{
		if( values_list_element == NULL )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: missing values list element: %d.",
			 function,
			 values_list_element_index );

			return( -1 );
		}
		if( values_list_element->value == NULL )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid values list element: %d - missing value.",
			 function,
			 values_list_element_index );

			return( -1 );
		}
		*block_descriptor = (libvshadow_block_descriptor_t *) values_list_element->value;

		if( ( relative_offset >= ( *block_descriptor )->relative_offset )
		 && ( relative_offset < ( ( *block_descriptor )->relative_offset + 0x4000 ) ) )
		{
			return( 1 );
		}
		else if( relative_offset < ( *block_descriptor )->relative_offset )
		{
			break;
		}
		values_list_element = values_list_element->next_element;

		if( node->number_of_sub_nodes != 0 )
		{
			if( sub_node == NULL )
			{
				libcerror_error_set(
				 error,
				 LIBCERROR_ERROR_DOMAIN_RUNTIME,
				 LIBCERROR_RUNTIME_ERROR_VALUE_MISSING,
				 "%s: missing sub node: %d.",
				 function,
				 sub_node_index );

				return( -1 );
			}
			sub_node = sub_node->next_node;

			sub_node_index++;
		}
	}
	*block_descriptor = NULL;

	if( node->number_of_sub_nodes == 0 )
	{
		return( 0 );
	}
	result = libvshadow_block_tree_node_get_descriptor_by_relative_offset(
		  sub_node,
		  relative_offset,
		  block_descriptor,
		  error );

	if( result == -1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve block descriptor in sub node: %d.",
		 function,
		 sub_node_index );

		return( -1 );
	}
	return( result );
}

/* Retrieves the block descriptor for the specific orignal offset
 * Returns 1 if a block decriptor was found, 0 if not or -1 on error
 */
int libvshadow_block_tree_get_range_by_original_offset(
     libvshadow_btree_t *block_tree,
     off64_t original_offset,
     off64_t *block_offset,
     size_t *block_size,
     uint32_t *block_flags,
     libcerror_error_t **error )
{
	libvshadow_block_descriptor_t *block_descriptor                   = NULL;
	libvshadow_block_descriptor_t *overlay_block_descriptor           = NULL;
	libvshadow_block_descriptor_t *overlay_expansion_block_descriptor = NULL;
	static char *function                                             = "libvshadow_block_tree_get_range_by_original_offset";
	size_t offset_correction                                          = 0;
	uint16_t next_overlay_block_offset                                = 0;
	uint16_t previous_overlay_block_offset                            = 0;
	int overlay_block_descriptor_index                                = 0;
	int overlay_expansion_block_descriptor_index                      = 0;
	int number_of_overlays                                            = 0;
	int number_of_overlay_expansions                                  = 0;
	int result                                                        = 0;

	if( block_tree == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid block tree.",
		 function );

		return( -1 );
	}
	if( block_offset == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid block offset.",
		 function );

		return( -1 );
	}
	if( block_size == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid block size.",
		 function );

		return( -1 );
	}
	if( block_flags == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid block flags.",
		 function );

		return( -1 );
	}
fprintf( stderr, "O: 0x%08" PRIx64 "\n",
 original_offset );
	result = libvshadow_block_tree_node_get_descriptor_by_original_offset(
		  block_tree->root_node,
		  original_offset,
		  &block_descriptor,
		  error );

	if( result == -1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve block descriptor in root node.",
		 function );

		return( -1 );
	}
	else if( result == 0 )
	{
		offset_correction = (size_t) ( original_offset % 0x4000 );

		*block_offset = original_offset;
		*block_size   = 0x4000 - offset_correction;
		*block_flags  = 0;

fprintf( stderr, "MISS: 0x%08" PRIx64 ", size: %" PRIzd "\n",
 *block_offset, *block_size );

		return( 0 );
	}
	if( block_descriptor == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_VALUE_MISSING,
		 "%s: missing block descriptor.",
		 function );

		return( -1 );
	}
	if( libvshadow_array_get_number_of_entries(
	     block_descriptor->overlays_array,
	     &number_of_overlays,
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve number of overlays from block descriptor.",
		 function );

		return( -1 );
	}
	previous_overlay_block_offset = 0;
	next_overlay_block_offset     = 0x4000;

	for( overlay_block_descriptor_index = 0;
	     overlay_block_descriptor_index < number_of_overlays;
	     overlay_block_descriptor_index++ )
	{
		if( libvshadow_array_get_entry_by_index(
		     block_descriptor->overlays_array,
		     overlay_block_descriptor_index,
		     (intptr_t **) &overlay_block_descriptor,
		     error ) != 1 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve overlay block descriptor: %d.",
			 function,
			 overlay_block_descriptor_index );

			return( -1 );
		}
		if( overlay_block_descriptor == NULL )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: missing overlay block descriptor: %d.",
			 function,
			 overlay_block_descriptor_index );

			return( -1 );
		}
		if( ( overlay_block_descriptor->flags & LIBVSHADOW_BLOCK_DESCRIPTOR_FLAGS_IS_OVERLAY ) == 0 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBCERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
			 "%s: unsupported overlay block descriptor flags.",
			 function );

			return( -1 );
		}
		*block_offset = overlay_block_descriptor->original_offset
			      + overlay_block_descriptor->block_offset;

		*block_size = overlay_block_descriptor->block_size;

		if( ( original_offset >= *block_offset )
		 && ( original_offset < (off64_t) ( *block_offset + *block_size ) ) )
		{
#if defined( HAVE_DEBUG_OUTPUT )
			if( libcnotify_verbose != 0 )
			{
				libcnotify_printf(
				 "%s: block overlay: %d offset: 0x%08" PRIx64 " size: %" PRIu16 "\n",
				 function,
				 overlay_block_descriptor_index,
				 *block_offset,
				 *block_size );
			}
#endif
			offset_correction = (size_t) ( original_offset - *block_offset );

			*block_offset = overlay_block_descriptor->offset
			              + overlay_block_descriptor->block_offset
			              + offset_correction;
			*block_size  -= offset_correction;
			*block_flags  = overlay_block_descriptor->flags;

fprintf( stderr, "OVERLAY: 0x%08" PRIx64 ", size: %" PRIzd " (0x%08" PRIx32 ")\n",
*block_offset, *block_size, overlay_block_descriptor->flags );

			return( 1 );
		}
		if( original_offset >= *block_offset )
		{
			if( ( overlay_block_descriptor->block_offset + overlay_block_descriptor->block_size ) > previous_overlay_block_offset )
			{
				previous_overlay_block_offset = overlay_block_descriptor->block_offset
				                              + overlay_block_descriptor->block_size;
			}
fprintf( stderr, "PREVIOUS OVERLAY (1): %" PRIu16 "\n", previous_overlay_block_offset );
		}
		if( original_offset < *block_offset )
		{
			if( overlay_block_descriptor->block_offset < next_overlay_block_offset )
			{
				next_overlay_block_offset = overlay_block_descriptor->block_offset;
			}
fprintf( stderr, "NEXT OVERLAY (1): %" PRIu16 "\n", next_overlay_block_offset );
		}
		if( libvshadow_array_get_number_of_entries(
		     overlay_block_descriptor->overlays_array,
		     &number_of_overlay_expansions,
		     error ) != 1 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve number of overlay expansions from overlay block descriptor: %d.",
			 function,
			 overlay_block_descriptor_index );

			return( -1 );
		}
		for( overlay_expansion_block_descriptor_index = 0;
		     overlay_expansion_block_descriptor_index < number_of_overlay_expansions;
		     overlay_expansion_block_descriptor_index++ )
		{
			if( libvshadow_array_get_entry_by_index(
			     overlay_block_descriptor->overlays_array,
			     overlay_expansion_block_descriptor_index,
			     (intptr_t **) &overlay_expansion_block_descriptor,
			     error ) != 1 )
			{
				libcerror_error_set(
				 error,
				 LIBCERROR_ERROR_DOMAIN_RUNTIME,
				 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
				 "%s: unable to retrieve expansion overlay block descriptor: %d.",
				 function,
				 overlay_expansion_block_descriptor_index );

				return( -1 );
			}
			if( overlay_expansion_block_descriptor == NULL )
			{
				libcerror_error_set(
				 error,
				 LIBCERROR_ERROR_DOMAIN_RUNTIME,
				 LIBCERROR_RUNTIME_ERROR_VALUE_MISSING,
				 "%s: missing overlay expansion block descriptor: %d.",
				 function,
				 overlay_expansion_block_descriptor_index );

				return( -1 );
			}
			if( ( overlay_expansion_block_descriptor->flags & LIBVSHADOW_BLOCK_DESCRIPTOR_FLAGS_IS_OVERLAY ) == 0 )
			{
				libcerror_error_set(
				 error,
				 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
				 LIBCERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
				 "%s: unsupported overlay expansion block descriptor flags.",
				 function );

				return( -1 );
			}
			*block_offset = overlay_expansion_block_descriptor->original_offset
				      + overlay_expansion_block_descriptor->block_offset;

			*block_size = overlay_expansion_block_descriptor->block_size;

			if( ( original_offset >= *block_offset )
			 && ( original_offset < (off64_t) ( *block_offset + *block_size ) ) )
			{
#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					libcnotify_printf(
					 "%s: block overlay expansion: %d offset: 0x%08" PRIx64 " size: %" PRIu16 "\n",
					 function,
					 overlay_expansion_block_descriptor_index,
					 *block_offset,
					 *block_size );
				}
#endif
				offset_correction = (size_t) ( original_offset - *block_offset );

/* TODO sanity check of overlay_expansion_block_descriptor->relative_offset == 0x00000001 ?*/
				*block_offset = overlay_expansion_block_descriptor->offset
				              + overlay_expansion_block_descriptor->block_offset
				              + offset_correction;
				*block_size  -= offset_correction;
				*block_flags  = overlay_expansion_block_descriptor->flags;

fprintf( stderr, "OVERLAY EXPANSION: 0x%08" PRIx64 ", size: %" PRIzd " (0x%08" PRIx32 ")\n",
*block_offset, *block_size, overlay_expansion_block_descriptor->flags );

				return( 1 );
			}
			if( original_offset >= *block_offset )
			{
				if( ( overlay_expansion_block_descriptor->block_offset + overlay_expansion_block_descriptor->block_size ) > previous_overlay_block_offset )
				{
					previous_overlay_block_offset = overlay_expansion_block_descriptor->block_offset
					                              + overlay_expansion_block_descriptor->block_size;
				}
fprintf( stderr, "PREVIOUS OVERLAY (2): %" PRIu16 "\n", previous_overlay_block_offset );
			}
			if( original_offset < *block_offset )
			{
				if( overlay_expansion_block_descriptor->block_offset < next_overlay_block_offset )
				{
					next_overlay_block_offset = overlay_expansion_block_descriptor->block_offset;
				}
fprintf( stderr, "NEXT OVERLAY (2): %" PRIu16 "\n", next_overlay_block_offset );
			}
		}
	}
	*block_offset = block_descriptor->original_offset
	              + block_descriptor->block_offset;

	*block_size   = block_descriptor->block_size;

	if( ( block_descriptor->flags & LIBVSHADOW_BLOCK_DESCRIPTOR_FLAGS_IS_OVERLAY ) != 0 )
	{
		if( original_offset >= *block_offset )
		{
			if( ( block_descriptor->block_offset + block_descriptor->block_size ) > previous_overlay_block_offset )
			{
				previous_overlay_block_offset = block_descriptor->block_offset
				                              + block_descriptor->block_size;
			}
fprintf( stderr, "PREVIOUS OVERLAY (3): %" PRIu16 "\n", previous_overlay_block_offset );
		}
		if( original_offset < *block_offset )
		{
			if( block_descriptor->block_offset < next_overlay_block_offset )
			{
				next_overlay_block_offset = block_descriptor->block_offset;
			}
fprintf( stderr, "NEXT OVERLAY (3): %" PRIu16 "\n", next_overlay_block_offset );
		}
	}
fprintf( stderr, "OVERLAY PREVIOUS: %" PRIu16 ", NEXT: %" PRIu16 "\n", previous_overlay_block_offset, next_overlay_block_offset );

	if( ( original_offset >= *block_offset )
	 && ( original_offset < (off64_t) ( *block_offset + *block_size ) ) )
	{
#if defined( HAVE_DEBUG_OUTPUT )
		if( libcnotify_verbose != 0 )
		{
			libcnotify_printf(
			 "%s: block offset: 0x%08" PRIx64 " size: %" PRIu16 "\n",
			 function,
			 *block_offset,
			 *block_size );
		}
#endif
		offset_correction = (size_t) ( original_offset - *block_offset );

		if( ( block_descriptor->flags & LIBVSHADOW_BLOCK_DESCRIPTOR_FLAGS_IS_FORWARDER ) == 0 )
		{
			*block_offset = block_descriptor->offset;
		}
		else
		{
			*block_offset = block_descriptor->relative_offset;
		}
		*block_offset += block_descriptor->block_offset + offset_correction;

		if( next_overlay_block_offset < ( block_descriptor->block_offset + block_descriptor->block_size ) )
		{
			*block_size = next_overlay_block_offset;
		}
		*block_size -= offset_correction;
		*block_flags = block_descriptor->flags;

fprintf( stderr, "HIT: 0x%08" PRIx64 ", size: %" PRIzd " (0x%08" PRIx32 ")\n",
 *block_offset, *block_size, block_descriptor->flags );

		result = 1;
	}
	else if( original_offset < *block_offset )
	{
#if defined( HAVE_DEBUG_OUTPUT )
		if( libcnotify_verbose != 0 )
		{
			libcnotify_printf(
			 "%s: before block offset: 0x%08" PRIx64 " size: %" PRIu16 "\n",
			 function,
			 block_descriptor->original_offset + previous_overlay_block_offset,
			 next_overlay_block_offset - previous_overlay_block_offset );
		}
#endif
		offset_correction = (size_t) ( original_offset - ( block_descriptor->original_offset + previous_overlay_block_offset ) );

		*block_offset = original_offset;
		*block_size   = (size_t) ( next_overlay_block_offset - previous_overlay_block_offset ) - offset_correction;
		*block_flags  = 0;

fprintf( stderr, "PRE: 0x%08" PRIx64 ", size: %" PRIzd "\n",
 *block_offset, *block_size );

		result = 0;
	}
	else
	{
#if defined( HAVE_DEBUG_OUTPUT )
		if( libcnotify_verbose != 0 )
		{
			libcnotify_printf(
			 "%s: after block offset: 0x%08" PRIx64 " size: %" PRIu16 "\n",
			 function,
			 *block_offset + *block_size,
			 (uint16_t) ( next_overlay_block_offset - ( block_descriptor->block_offset + block_descriptor->block_size ) ) );
		}
#endif
		offset_correction = (size_t) ( original_offset - ( *block_offset + *block_size ) );

		*block_offset = original_offset;
		*block_size   = (size_t) ( next_overlay_block_offset - ( block_descriptor->block_offset + block_descriptor->block_size ) )
		              - offset_correction;
		*block_flags  = 0;

fprintf( stderr, "POST: 0x%08" PRIx64 ", size: %" PRIzd "\n",
 *block_offset, *block_size );

		result = 0;
	}
	return( result );
}

/* Inserts a block descriptor in the block tree
 * Returns 1 if successful or -1 on error
 */
int libvshadow_block_tree_insert(
     libvshadow_btree_t *forward_block_tree,
     libvshadow_btree_t *reverse_block_tree,
     libvshadow_block_descriptor_t *block_descriptor,
     libcerror_error_t **error )
{
	libvshadow_block_descriptor_t *existing_block_descriptor = NULL;
	libvshadow_list_element_t *existing_list_element         = NULL;
	static char *function                                    = "libvshadow_block_tree_insert";
	int overlays_array_index                                 = 0;
	int result                                               = 0;

	if( forward_block_tree == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid forward block tree.",
		 function );

		return( -1 );
	}
	if( reverse_block_tree == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid reverse block tree.",
		 function );

		return( -1 );
	}
	if( block_descriptor == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid block descriptor.",
		 function );

		return( -1 );
	}
	result = libvshadow_block_tree_node_get_descriptor_by_relative_offset(
		  reverse_block_tree->root_node,
		  block_descriptor->original_offset,
		  &existing_block_descriptor,
		  error );

	if( result == -1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve block descriptor in root node.",
		 function );

		return( -1 );
	}
	else if( result != 0 )
	{
fprintf( stderr, "REVERSE FOUND\n" );
	}
	result = libvshadow_btree_insert_value(
		  forward_block_tree,
	          &( block_descriptor->index ),
		  (intptr_t *) block_descriptor,
		  (int (*)(intptr_t *, intptr_t *, libcerror_error_t **)) &libvshadow_block_descriptor_compare_by_original_offset,
		  &existing_list_element,
		  error );

	if( result == -1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_APPEND_FAILED,
		 "%s: unable to insert block descriptor in forward block tree.",
		 function );

		return( -1 );
	}
	else if( result == 0 )
	{
		if( existing_list_element == NULL )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: missing existing list element.",
			 function );

			return( -1 );
		}
		if( existing_list_element->value == NULL )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid existing list element - missing value.",
			 function );

			return( -1 );
		}
		existing_block_descriptor = (libvshadow_block_descriptor_t *) existing_list_element->value;

		if( ( block_descriptor->flags & LIBVSHADOW_BLOCK_DESCRIPTOR_FLAGS_IS_OVERLAY ) != 0 )
		{
			if( block_descriptor->relative_offset == 0x00000001UL )
			{
				if( block_descriptor->offset != existing_block_descriptor->offset )
				{
					if( existing_list_element->value == NULL )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_INPUT,
						 LIBCERROR_INPUT_ERROR_VALUE_MISMATCH,
						 "%s: mismatch in block descriptor store offset .",
						 function );

						return( -1 );
					}
				}
/* TODO */
fprintf( stderr, "OVERLAY EXPANSION: 0x%08" PRIx32 "\n",
 existing_block_descriptor->flags );
			}
			else
			{
/* TODO */
fprintf( stderr, "OVERLAY: 0x%08" PRIx32 "\n",
 existing_block_descriptor->flags );
			}
			if( libvshadow_array_append_entry(
			     existing_block_descriptor->overlays_array,
			     &overlays_array_index,
			     (intptr_t *) block_descriptor,
			     error ) != 1 )
			{
				libcerror_error_set(
				 error,
				 LIBCERROR_ERROR_DOMAIN_RUNTIME,
				 LIBCERROR_RUNTIME_ERROR_APPEND_FAILED,
				 "%s: unable to append block descriptor to overlays array.",
				 function );

				return( -1 );
			}
		}
		else
		{
/* TODO */
fprintf( stderr, "REPLACE: 0x%08" PRIx32 "\n",
 existing_block_descriptor->flags );

			if( libvshadow_array_set_entry_by_index(
			     forward_block_tree->values_array,
			     existing_block_descriptor->index,
			     (intptr_t *) block_descriptor,
			     error ) != 1 )
			{
				libcerror_error_set(
				 error,
				 LIBCERROR_ERROR_DOMAIN_RUNTIME,
				 LIBCERROR_RUNTIME_ERROR_APPEND_FAILED,
				 "%s: unable to set block descriptor: %d in btree values array.",
				 function,
				 existing_block_descriptor->index );

				return( -1 );
			}
			block_descriptor->index = existing_block_descriptor->index;

			existing_block_descriptor->index = -1;

			existing_list_element->value = (intptr_t *) block_descriptor;

			if( ( existing_block_descriptor->flags & LIBVSHADOW_BLOCK_DESCRIPTOR_FLAGS_IS_OVERLAY ) != 0 )
			{
/* TODO */
fprintf( stderr, "APPENDING EXISTING: 0x%08" PRIx32 "\n",
 existing_block_descriptor->flags );
				if( libvshadow_array_append_entry(
				     block_descriptor->overlays_array,
				     &overlays_array_index,
				     (intptr_t *) existing_block_descriptor,
				     error ) != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_APPEND_FAILED,
					 "%s: unable to append existing block descriptor to overlays array.",
					 function );

					return( -1 );
				}
			}
			else
			{
/* TODO */
fprintf( stderr, "DISCARDING EXISTING: 0x%08" PRIx32 "\n",
 existing_block_descriptor->flags );
			}
		}
	}
	if( ( block_descriptor->flags & LIBVSHADOW_BLOCK_DESCRIPTOR_FLAGS_IS_FORWARDER ) != 0 )
	{
		result = libvshadow_btree_insert_value(
			  reverse_block_tree,
			  &( block_descriptor->index ),
			  (intptr_t *) block_descriptor,
			  (int (*)(intptr_t *, intptr_t *, libcerror_error_t **)) &libvshadow_block_descriptor_compare_by_relative_offset,
			  &existing_list_element,
			  error );

		if( result == -1 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_APPEND_FAILED,
			 "%s: unable to insert block descriptor in reverse block tree.",
			 function );

			return( -1 );
		}
		else if( result == 0 )
		{
		}
	}
	return( 1 );
}

