/*
 * B-tree type functions
 *
 * Copyright (c) 2011-2012, Joachim Metz <jbmetz@users.sourceforge.net>
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <memory.h>
#include <types.h>

#include "libvshadow_btree_type.h"
#include "libvshadow_libcerror.h"
#include "libvshadow_list_type.h"
#include "libvshadow_tree_type.h"

/* Creates a tree
 * Returns 1 if successful or -1 on error
 */
int libvshadow_btree_initialize(
     libvshadow_btree_t **tree,
     int maximum_number_of_values,
     libcerror_error_t **error )
{
	static char *function = "libvshadow_btree_initialize";

	if( tree == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid tree.",
		 function );

		return( -1 );
	}
	if( maximum_number_of_values <= 0 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_VALUE_OUT_OF_BOUNDS,
		 "%s: invalid maximum number of values value out of bounds.",
		 function );

		return( -1 );
	}
	if( *tree != NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid tree value already set.",
		 function );

		return( -1 );
	}
	*tree = memory_allocate_structure(
	         libvshadow_btree_t );

	if( *tree == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_MEMORY,
		 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
		 "%s: unable to create tree.",
		 function );

		goto on_error;
	}
	if( memory_set(
	     *tree,
	     0,
	     sizeof( libvshadow_btree_t ) ) == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_MEMORY,
		 LIBCERROR_MEMORY_ERROR_SET_FAILED,
		 "%s: unable to clear tree.",
		 function );

		memory_free(
		 *tree );

		*tree = NULL;

		return( -1 );
	}
	if( libvshadow_array_initialize(
	     &( ( *tree )->values_array ),
	     0,
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create values array.",
		 function );

		goto on_error;
	}
	if( libvshadow_tree_node_initialize(
	     &( ( *tree )->root_node ),
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create root node.",
		 function );

		goto on_error;
	}
	( *tree )->maximum_number_of_values = maximum_number_of_values;

	return( 1 );

on_error:
	if( *tree != NULL )
	{
		if( ( *tree )->values_array != NULL )
		{
			libvshadow_array_free(
			 &( ( *tree )->values_array ),
			 NULL,
			 NULL );
		}
		memory_free(
		 *tree );

		*tree = NULL;
	}
	return( -1 );
}

/* Frees a tree and its nodes
 * Uses the value_free_function to free the value
 * Returns 1 if successful or -1 on error
 */
int libvshadow_btree_free(
     libvshadow_btree_t **tree,
     int (*value_free_function)(
            intptr_t **value,
            libcerror_error_t **error ),
     libcerror_error_t **error )
{
	static char *function = "libvshadow_btree_free";
	int result            = 1;

	if( tree == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid tree.",
		 function );

		return( -1 );
	}
	if( *tree != NULL )
	{
		if( libvshadow_tree_node_free(
		     &( ( *tree )->root_node ),
		     (int (*)(intptr_t **, libcerror_error_t **)) &libvshadow_btree_free_values_list,
		     error ) != 1 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_FINALIZE_FAILED,
			 "%s: unable to free root node.",
			 function );

			result = -1;
		}
		if( libvshadow_array_free(
		     &( ( *tree )->values_array ),
		     value_free_function,
		     error ) != 1 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_FINALIZE_FAILED,
			 "%s: unable to free values array.",
			 function );

			result = -1;
		}
		memory_free(
		 *tree );

		*tree = NULL;
	}
	return( result );
}

/* Frees a values list
 * Returns 1 if successful or -1 on error
 */
int libvshadow_btree_free_values_list(
     libvshadow_list_t **values_list,
     libcerror_error_t **error )
{
	static char *function = "libvshadow_btree_free_values_list";
	int result            = 1;

	if( values_list == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid values list.",
		 function );

		return( -1 );
	}
	if( libvshadow_list_free(
	     values_list,
	     NULL,
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_FINALIZE_FAILED,
		 "%s: unable to free values list.",
		 function );

		result = -1;
	}
	return( result );
}

/* TODO clone function */

/* Retrieves the number of values in the tree
 * Returns 1 if successful or -1 on error
 */
int libvshadow_btree_get_number_of_values(
     libvshadow_btree_t *tree,
     int *number_of_values,
     libcerror_error_t **error )
{
	static char *function = "libvshadow_btree_get_number_of_values";

	if( tree == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid tree.",
		 function );

		return( -1 );
	}
	if( libvshadow_array_get_number_of_entries(
	     tree->values_array,
	     number_of_values,
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve number of values array entries.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves a specific value
 * Returns 1 if successful or -1 on error
 */
int libvshadow_btree_get_value_by_index(
     libvshadow_btree_t *tree,
     int value_index,
     intptr_t **value,
     libcerror_error_t **error )
{
	static char *function = "libvshadow_btree_get_value_by_index";

	if( tree == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid tree.",
		 function );

		return( -1 );
	}
	if( libvshadow_array_get_entry_by_index(
	     tree->values_array,
	     value_index,
	     value,
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve value: %d from array.",
		 function,
		 value_index );

		return( -1 );
	}
	return( 1 );
}

/* Retrieves the upper node for the specific value
 * Returns 1 if successful, 0 if the value already exists or -1 on error
 */
int libvshadow_btree_node_get_upper_node_by_value(
     libvshadow_tree_node_t *node,
     intptr_t *value,
     int (*value_compare_function)(
            intptr_t *first_value,
            intptr_t *second_value,
            libcerror_error_t **error ),
     libvshadow_tree_node_t **upper_node,
     libvshadow_list_element_t **upper_list_element,
     libcerror_error_t **error )
{
	libvshadow_list_element_t *values_list_element = NULL;
	libvshadow_tree_node_t *sub_node               = NULL;
	static char *function                          = "libvshadow_btree_node_get_upper_node_by_value";
	int number_of_values_list_elements             = 0;
	int result                                     = 0;
	int sub_node_index                             = 0;
	int values_list_element_index                  = 0;

	if( node == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid node.",
		 function );

		return( -1 );
	}
	if( value == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid value.",
		 function );

		return( -1 );
	}
	if( value_compare_function == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid value compare function.",
		 function );

		return( -1 );
	}
	if( upper_node == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid upper node.",
		 function );

		return( -1 );
	}
	if( upper_list_element == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid upper list element.",
		 function );

		return( -1 );
	}
	*upper_list_element = NULL;

	if( node->value != NULL )
	{
		if( libvshadow_list_get_number_of_elements(
		     (libvshadow_list_t *) node->value,
		     &number_of_values_list_elements,
		     error ) != 1 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to retrieve number of values list elements.",
			 function );

			return( -1 );
		}
	}
	if( number_of_values_list_elements == 0 )
	{
		*upper_node = node;

		return( 1 );
	}
	if( libvshadow_list_get_element_by_index(
	     (libvshadow_list_t *) node->value,
	     0,
	     &values_list_element,
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve values list element: 0.",
		 function );

		return( -1 );
	}
	if( node->number_of_sub_nodes != 0 )
	{
		if( ( number_of_values_list_elements + 1 ) != node->number_of_sub_nodes )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBCERROR_ARGUMENT_ERROR_VALUE_OUT_OF_BOUNDS,
			 "%s: invalid number of values list elements value out of bounds.",
			 function );

			return( -1 );
		}
		sub_node = node->first_sub_node;
	}
	for( values_list_element_index = 0;
	     values_list_element_index < number_of_values_list_elements;
	     values_list_element_index++ )
	{
		if( values_list_element == NULL )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: missing values list element: %d.",
			 function,
			 values_list_element_index );

			return( -1 );
		}
		if( values_list_element->value == NULL )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid values list element: %d - missing value.",
			 function,
			 values_list_element_index );

			return( -1 );
		}
		result = value_compare_function(
			  value,
			  values_list_element->value,
			  error );

		if( result == -1 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
			 "%s: unable to compare values list element: %d value.",
			 function,
			 values_list_element_index );

			return( -1 );
		}
		else if( result == LIBVSHADOW_BTREE_NODE_COMPARE_EQUAL )
		{
			*upper_node         = node;
			*upper_list_element = values_list_element;

			return( 0 );
		}
		else if( result == LIBVSHADOW_BTREE_NODE_COMPARE_LESS )
		{
			break;
		}
		else if( result != LIBVSHADOW_BTREE_NODE_COMPARE_GREATER )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
			 LIBCERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
			 "%s: unsupported value compare function return value: %d.",
			 function,
			 result );

			return( -1 );
		}
		values_list_element = values_list_element->next_element;

		if( node->number_of_sub_nodes != 0 )
		{
			if( sub_node == NULL )
			{
				libcerror_error_set(
				 error,
				 LIBCERROR_ERROR_DOMAIN_RUNTIME,
				 LIBCERROR_RUNTIME_ERROR_VALUE_MISSING,
				 "%s: missing sub node: %d.",
				 function,
				 sub_node_index );

				return( -1 );
			}
			sub_node = sub_node->next_node;

			sub_node_index++;
		}
	}
	if( node->number_of_sub_nodes == 0 )
	{
		*upper_node = node;

		return( 1 );
	}
	result = libvshadow_btree_node_get_upper_node_by_value(
		  sub_node,
		  value,
		  value_compare_function,
		  upper_node,
		  upper_list_element,
		  error );

	if( result == -1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve upper node in sub node: %d.",
		 function,
		 sub_node_index );

		return( -1 );
	}
	if( *upper_node == NULL )
	{
		*upper_node = sub_node;
	}
	return( result );
}

/* Appends a value into a tree node
 * Returns 1 if successful or -1 on error
 */
int libvshadow_btree_node_append_value(
     libvshadow_tree_node_t *node,
     intptr_t *value,
     libcerror_error_t **error )
{
	static char *function = "libvshadow_btree_node_append_value";

	if( node == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid node.",
		 function );

		return( -1 );
	}
	if( node->value == NULL )
	{
		if( libvshadow_list_initialize(
		     (libvshadow_list_t **) &( node->value ),
		     error ) != 1 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to create values list.",
			 function );

			return( -1 );
		}
	}
	if( libvshadow_list_append_value(
	     (libvshadow_list_t *) node->value,
	     value,
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_APPEND_FAILED,
		 "%s: unable to append value to values list.",
		 function );

		return( -1 );
	}
	return( 1 );
}

/* Inserts a value into a tree node
 * Returns 1 if successful, 0 if the value already exists or -1 on error
 */
int libvshadow_btree_node_insert_value(
     libvshadow_tree_node_t *node,
     intptr_t *value,
     int (*value_compare_function)(
            intptr_t *first_value,
            intptr_t *second_value,
            libcerror_error_t **error ),
     libcerror_error_t **error )
{
	static char *function = "libvshadow_btree_node_insert_value";
	int result            = 0;

	if( node == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid node.",
		 function );

		return( -1 );
	}
	if( node->value == NULL )
	{
		if( libvshadow_list_initialize(
		     (libvshadow_list_t **) &( node->value ),
		     error ) != 1 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to create values list.",
			 function );

			return( -1 );
		}
	}
	result = libvshadow_list_insert_value(
	          (libvshadow_list_t *) node->value,
	          value,
	          value_compare_function,
	          LIBVSHADOW_LIST_INSERT_FLAG_UNIQUE_ENTRIES,
	          error );

	if( result == -1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_APPEND_FAILED,
		 "%s: unable to insert value in values list.",
		 function );

		return( -1 );
	}
	return( result );
}

/* Splits the node
 * Returns 1 if successful or -1 on error
 */
int libvshadow_btree_split_node(
     libvshadow_tree_node_t *node,
     libcerror_error_t **error )
{
	libvshadow_list_t *values_list                 = NULL;
	libvshadow_list_element_t *values_list_element = NULL;
	libvshadow_tree_node_t *left_sub_node          = NULL;
	libvshadow_tree_node_t *right_sub_node         = NULL;
	static char *function                          = "libvshadow_btree_split_node";
	int number_of_values_list_elements             = 0;
	int split_values_list_element_index            = 0;
	int values_list_element_index                  = 0;

	if( node == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid node.",
		 function );

		return( -1 );
	}
	if( node->number_of_sub_nodes != 0 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
		 "%s: cannot split node with sub nodes.",
		 function );

		return( -1 );
	}
	if( libvshadow_list_get_number_of_elements(
	     (libvshadow_list_t *) node->value,
	     &number_of_values_list_elements,
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve number of values list elements.",
		 function );

		goto on_error;
	}
	split_values_list_element_index = number_of_values_list_elements / 2;

	if( ( number_of_values_list_elements % 2 ) != 0 )
	{
		split_values_list_element_index += 1;
	}
	/* Keep the values list in case of an error
	 */
	values_list = (libvshadow_list_t *) node->value;
	node->value = NULL;

	if( libvshadow_tree_node_initialize(
	     &left_sub_node,
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create left sub node.",
		 function );

		goto on_error;
	}
	if( libvshadow_tree_node_append_node(
	     node,
	     left_sub_node,
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_APPEND_FAILED,
		 "%s: unable to append left sub node to node.",
		 function );

		goto on_error;
	}
	if( libvshadow_tree_node_initialize(
	     &right_sub_node,
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create right sub node.",
		 function );

		goto on_error;
	}
	if( libvshadow_tree_node_append_node(
	     node,
	     right_sub_node,
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_APPEND_FAILED,
		 "%s: unable to append right sub node to node.",
		 function );

		goto on_error;
	}
	if( libvshadow_list_get_element_by_index(
	     values_list,
	     0,
	     &values_list_element,
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve values list element: 0.",
		 function );

		goto on_error;
	}
	for( values_list_element_index = 0;
	     values_list_element_index < number_of_values_list_elements;
	     values_list_element_index++ )
	{
		if( values_list_element == NULL )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: missing values list element: %d.",
			 function,
			 values_list_element_index );

			goto on_error;
		}
		if( values_list_element->value == NULL )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_VALUE_MISSING,
			 "%s: invalid values list element: %d - missing value.",
			 function,
			 values_list_element_index );

			goto on_error;
		}
		if( values_list_element_index < split_values_list_element_index )
		{
			if( libvshadow_btree_node_append_value(
			     left_sub_node,
			     values_list_element->value,
			     error ) != 1 )
			{
				libcerror_error_set(
				 error,
				 LIBCERROR_ERROR_DOMAIN_RUNTIME,
				 LIBCERROR_RUNTIME_ERROR_APPEND_FAILED,
				 "%s: unable to append value to left sub node.",
				 function );

				goto on_error;
			}
		}
		else if( values_list_element_index > split_values_list_element_index )
		{
			if( libvshadow_btree_node_append_value(
			     right_sub_node,
			     values_list_element->value,
			     error ) != 1 )
			{
				libcerror_error_set(
				 error,
				 LIBCERROR_ERROR_DOMAIN_RUNTIME,
				 LIBCERROR_RUNTIME_ERROR_APPEND_FAILED,
				 "%s: unable to append value to right sub node.",
				 function );

				goto on_error;
			}
		}
		else
		{
			if( libvshadow_btree_node_append_value(
			     node,
			     values_list_element->value,
			     error ) != 1 )
			{
				libcerror_error_set(
				 error,
				 LIBCERROR_ERROR_DOMAIN_RUNTIME,
				 LIBCERROR_RUNTIME_ERROR_APPEND_FAILED,
				 "%s: unable to append value to node.",
				 function );

				goto on_error;
			}
		}
		values_list_element = values_list_element->next_element;
	}
	if( libvshadow_list_free(
	     &values_list,
	     NULL,
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_FINALIZE_FAILED,
		 "%s: unable to free values list.",
		 function );

		return( -1 );
	}
	return( 1 );

on_error:
	if( values_list != NULL )
	{
		if( node->value != NULL )
		{
			libvshadow_list_free(
			 (libvshadow_list_t **) &( node->value ),
			 NULL,
			 NULL );
		}
		node->value = (intptr_t *) values_list;
	}
	if( right_sub_node != NULL )
	{
		libvshadow_tree_node_remove_node(
		 node,
		 right_sub_node,
		 NULL );
		libvshadow_tree_node_free(
		 &right_sub_node,
		 NULL,
		 NULL );
	}
	if( left_sub_node != NULL )
	{
		libvshadow_tree_node_remove_node(
		 node,
		 left_sub_node,
		 NULL );
		libvshadow_tree_node_free(
		 &left_sub_node,
		 NULL,
		 NULL );
	}
	return( -1 );
}

/* Inserts a value into a tree
 * Returns 1 if successful, 0 if the value already exists or -1 on error
 */
int libvshadow_btree_insert_value(
     libvshadow_btree_t *tree,
     int *value_index,
     intptr_t *value,
     int (*value_compare_function)(
            intptr_t *first_value,
            intptr_t *second_value,
            libcerror_error_t **error ),
     libvshadow_list_element_t **existing_list_element,
     libcerror_error_t **error )
{
	libvshadow_tree_node_t *upper_node = NULL;
	static char *function              = "libvshadow_btree_insert_value";
	int number_of_values               = 0;
	int result                         = 0;

	if( tree == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid tree.",
		 function );

		return( -1 );
	}
	if( value_index == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid value index.",
		 function );

		return( -1 );
	}
	if( existing_list_element == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid existing list element.",
		 function );

		return( -1 );
	}
	result = libvshadow_btree_node_get_upper_node_by_value(
	          tree->root_node,
	          value,
	          value_compare_function,
	          &upper_node,
	          existing_list_element,
	          error );

	if( result == -1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve upper node in root node.",
		 function );

		return( -1 );
	}
	else if( result == 0 )
	{
		return( 0 );
	}
	if( libvshadow_btree_node_insert_value(
	     upper_node,
	     value,
	     value_compare_function,
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_APPEND_FAILED,
		 "%s: unable to insert value in upper node.",
		 function );

		return( -1 );
	}
/* TODO */
	if( libvshadow_list_get_number_of_elements(
	     (libvshadow_list_t *) upper_node->value,
	     &number_of_values,
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
		 "%s: unable to retrieve number of sub nodes.",
		 function );

		return( -1 );
	}
	if( number_of_values >= tree->maximum_number_of_values )
	{
		if( libvshadow_btree_split_node(
		     upper_node,
		     error ) != 1 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_APPEND_FAILED,
			 "%s: unable to split upper node.",
			 function );

			return( -1 );
		}
/* TODO do merge of upper node with its parent node */
/* TODO loop until number_of_values < tree->maximum_number_of_values */
	}
	if( libvshadow_array_append_entry(
	     tree->values_array,
	     value_index,
	     value,
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_APPEND_FAILED,
		 "%s: unable to append value to values array.",
		 function );

		return( -1 );
	}
	return( 1 );
}

